from collections import defaultdict

import numpy as np
import panel as pn
import param

pn.extension(
    "tabulator",
    css_files=["https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.4.2/css/all.min.css"],
)


class AnnotatorTable(pn.viewable.Viewer):
    annotator = param.Parameter(allow_refs=False)
    tabulator = param.Parameter(allow_refs=False)
    dataframe = param.DataFrame()

    _updating = False

    def __init__(self, annotator, **params):
        super().__init__(annotator=annotator, **params)
        annotator.snapshot()
        self._create_tabulator()

    def _create_tabulator(self):
        def inner(event, annotator=self.annotator):
            return annotator.df

        def on_edit(event):
            row = self.tabulator.value.iloc[event.row]

            # Extracting specs and fields from row
            spec_dct, field_dct = defaultdict(list), {}
            for k, v in row.items():
                if "[" in k:
                    k = k.split("[")[1][:-1]  # Getting the spec name
                    spec_dct[k].append(v)
                else:
                    field_dct[k] = v

            self.annotator.annotation_table.update_annotation_region(spec_dct, row.name)
            self.annotator.update_annotation_fields(row.name, **field_dct)
            self.annotator.refresh(clear=True)

            # So it is still reactive, as editing overwrites the table
            self.tabulator.value = pn.bind(inner, self.annotator)

        def on_click(event):
            if event.column != "delete":
                return
            index = self.tabulator.value.iloc[event.row].name
            self.annotator.delete_annotation(index)

        def new_style(row):
            changed = [e["id"] for e in self.annotator.annotation_table._edits]
            color = "darkgray" if row.name in changed else "inherit"
            return [f"color: {color}"] * len(row)

        self.tabulator = pn.widgets.Tabulator(
            value=pn.bind(inner, self.annotator),
            buttons={"delete": '<i class="fa fa-trash"></i>'},
            show_index=False,
            selectable=True,
        )
        self.tabulator.on_edit(on_edit)
        self.tabulator.on_click(on_click)
        self.tabulator.style.apply(new_style, axis=1)

        def on_commit(event):
            self.tabulator.param.trigger("value")
            # So it is still reactive, as triggering the value overwrites the table
            self.tabulator.value = pn.bind(inner, self.annotator)

        self.annotator.on_commit(on_commit)

    @param.depends("tabulator.selection", watch=True)
    def _select_table_to_plot(self):
        if self._updating:
            return
        try:
            self._updating = True
            self.annotator.selected_indices = list(
                self.tabulator.value.iloc[self.tabulator.selection].index
            )
        except IndexError:
            pass  # when we delete we select and get an index error if it is the last
        finally:
            self._updating = False

    @param.depends("annotator.selected_indices", watch=True)
    def _select_plot_to_table(self):
        if self._updating:
            return
        try:
            self._updating = True
            # Likely better way to get this mapping
            mask = self.tabulator.value.index.isin(self.annotator.selected_indices)
            self.tabulator.selection = list(map(int, np.where(mask)[0]))

        finally:
            self._updating = False

    def clear(self):
        self.tabulator.selection = []
        self.tabulator.param.trigger("value")

    def __panel__(self):
        return self.tabulator



from __future__ import annotations

import numpy as np
import pandas as pd
import panel as pn

from holonote.app import AnnotatorTable, PanelWidgets


def test_panel_app(annotator_range1d):
    w = PanelWidgets(annotator_range1d)
    assert isinstance(w.fields_widgets, pn.Column)
    assert isinstance(w.tool_widgets, pn.Row)


def test_as_popup(annotator_range1d):
    w = PanelWidgets(annotator_range1d, as_popup=True)
    assert not w._layout.visible
    for display in w.annotator._displays.values():
        assert display._edit_streams[0].popup
        assert display._tap_stream.popup
    assert w.__panel__().visible


def test_tabulator(annotator_range1d):
    t = AnnotatorTable(annotator_range1d)
    assert isinstance(t.tabulator, pn.widgets.Tabulator)

    annotator_range1d.set_regions(TIME=(np.datetime64("2022-06-06"), np.datetime64("2022-06-08")))
    annotator_range1d.add_annotation(description="A test annotation!")
    assert len(t.tabulator.value) == 1
    assert t.tabulator.value.iloc[0, 0] == pd.Timestamp("2022-06-06")
    assert t.tabulator.value.iloc[0, 1] == pd.Timestamp("2022-06-08")
    assert t.tabulator.value.iloc[0, 2] == "A test annotation!"
    assert "darkgray" in t.tabulator.style.to_html()

    annotator_range1d.commit(return_commits=True)
    assert "darkgray" not in t.tabulator.style.to_html()



import holoviews as hv

bk_renderer = hv.renderer("bokeh")


def _get_display(annotator, kdims=None) -> hv.Element:
    if kdims is None:
        kdims = next(iter(annotator._displays))
    kdims = (kdims,) if isinstance(kdims, str) else tuple(kdims)
    disp = annotator.get_display(*kdims)
    bk_renderer.get_plot(disp.element)  # Trigger rendering
    return disp


def get_editor(annotator, element_type, kdims=None):
    el = _get_display(annotator, kdims).region_editor()
    for e in el.last.traverse():
        if isinstance(e, element_type):
            return e


def get_editor_data(annotator, element_type, kdims=None):
    el = get_editor(annotator, element_type, kdims)
    return getattr(el, "data", None)


def get_indicator(annotator, element_type, kdims=None):
    si = _get_display(annotator, kdims).indicators().last
    yield from si.data.values()


def get_indicator_data(annotator, element_type, kdims=None):
    for el in get_indicator(annotator, element_type, kdims):
        yield el.data


def get_display_data_from_plot(plot, element_type, kdims):
    display_data = None

    for el in plot.traverse():
        if isinstance(el, element_type):
            display_data = getattr(el, "data", None)
            break

    if display_data is not None:
        cols = [f"start[{dim}]" for dim in kdims] + [f"end[{dim}]" for dim in kdims]
        display_data = display_data.iloc[0][cols].to_list()

    return display_data



import holoviews as hv

hv.extension("bokeh")


class TestPoint2D:
    def test_get_indices_by_position_exact(self, annotator_point2d):
        x, y = 0.5, 0.3
        description = "A test annotation!"
        annotator_point2d.set_regions(x=x, y=y)
        annotator_point2d.add_annotation(description=description)
        display = annotator_point2d.get_display("x", "y")
        indices = display.get_indices_by_position(x=x, y=y)
        assert len(indices) == 1

    def test_get_indices_by_position_nearest_2d_point_threshold(self, annotator_point2d):
        x, y = 0.5, 0.3
        annotator_point2d.get_display("x", "y").nearest_2d_point_threshold = 1
        description = "A test annotation!"
        annotator_point2d.set_regions(x=x, y=y)
        annotator_point2d.add_annotation(description=description)
        display = annotator_point2d.get_display("x", "y")
        indices = display.get_indices_by_position(x=x + 1.5, y=y + 1.5)
        assert len(indices) == 0

        display.nearest_2d_point_threshold = 5
        indices = display.get_indices_by_position(x=x + 0.5, y=y + 0.5)
        assert len(indices) == 1

    def test_get_indices_by_position_nearest(self, annotator_point2d):
        x, y = 0.5, 0.3
        description = "A test annotation!"
        annotator_point2d.set_regions(x=x, y=y)
        annotator_point2d.add_annotation(description=description)
        display = annotator_point2d.get_display("x", "y")
        indices = display.get_indices_by_position(x=x + 1.5, y=y + 1.5)
        assert len(indices) == 1

        display.nearest_2d_point_threshold = 5
        indices = display.get_indices_by_position(x=x + 0.5, y=y + 0.5)
        assert len(indices) == 1

    def test_get_indices_by_position_empty(self, annotator_point2d):
        display = annotator_point2d.get_display("x", "y")
        indices = display.get_indices_by_position(x=0.5, y=0.3)
        assert len(indices) == 0

    def test_get_indices_by_position_no_position(self, annotator_point2d):
        display = annotator_point2d.get_display("x", "y")
        indices = display.get_indices_by_position(x=None, y=None)
        assert len(indices) == 0

    def test_get_indices_by_position_multi_choice(self, annotator_point2d):
        x, y = 0.5, 0.3
        description = "A test annotation!"
        annotator_point2d.set_regions(x=x, y=y)
        annotator_point2d.add_annotation(description=description)

        x2, y2 = 0.51, 0.31
        description = "A test annotation!"
        annotator_point2d.set_regions(x=x2, y=y2)
        annotator_point2d.add_annotation(description=description)

        display = annotator_point2d.get_display("x", "y")
        display.nearest_2d_point_threshold = 1000

        indices = display.get_indices_by_position(x=x, y=y)
        assert len(indices) == 1
